/*
 * OZ711MS1 OZ711MP1 linux driver
 * O2Micro Inc.
 * toby.yang@o2micro.com
 * 2005/4/17
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#ifndef _SCSI_BASIC_H
#define _SCSI_BASIC_H
/*
 * Inquiry Data Format
 */
typedef struct {
   UCHAR   DeviceType;
   UCHAR   DeviceTypeQualifier;
   UCHAR   Version;
   UCHAR   ResponseDataFormat;
   UCHAR   AdditionalLength;
   UCHAR   Reserved;
   UCHAR   Flags[2];
   UCHAR   VendorId[8];
   UCHAR   ProductId[16];
   UCHAR   ProductRevisionLevel[4];
   UCHAR   Reserved2;                                  /* Provides NULL terminator to name */
}INQUIRY_DATA,*PINQUIRY_DATA;

typedef struct {
	UCHAR	ResponseCode;
	UCHAR Reserved;
	UCHAR SenseKey;
	UCHAR Infomation[4];
	UCHAR AdditionalSenseLength;
	UCHAR CommandSpecInfo[4];
	UCHAR AdditionalSenseCode;
	UCHAR AdditionalSenseCodeQual;
	UCHAR FieldReplaceable;
	UCHAR SenseKeySpec[3];
}SENSE_DATA,*PSENSE_DATA;
/*
 * Sense Data Format - Page 3
 */
typedef struct {
   UCHAR  PageCode;
   UCHAR  PageLength;
   USHORT TracksPerZone;
   USHORT AltSectorsPerZone;
   USHORT AltTracksPerZone;
   USHORT AltTracksPerVolume;
   USHORT SectorsPerTrack;
   USHORT BytesPerSector;
   USHORT Interleave;
   USHORT TrackSkew;
   USHORT CylinderSkew;
   UCHAR  flags;
   UCHAR  reserved[3];
} MODE_PAGE3;
/*
 * Sense Data Format - Page 4
 */
typedef struct {
   UCHAR  PageCode;
   UCHAR  PageLength;
   USHORT CylindersHigh;
   UCHAR  CylindersLow;
   UCHAR  Heads;
   USHORT WritePrecompHigh;
   UCHAR  WritePrecompLow;
   USHORT ReducedWriteCurrentHigh;
   UCHAR  ReducedWriteCurrentLow;
   USHORT StepRate;
   USHORT LandingZoneHigh;
   UCHAR  LandingZoneLow;
   UCHAR  flags;
   UCHAR  RotationalOffset;
   UCHAR  Reserved;
   USHORT MediumRotationRate;
   UCHAR  Reserved2[2];
} MODE_PAGE4;

typedef struct {
    UCHAR PageCode;
    UCHAR PageLength;
    UCHAR Flags;
    UCHAR RetensionPriority;
    UCHAR DisablePrefetchTransfer[2];
    UCHAR MinimumPrefetch[2];
    UCHAR MaximumPrefetch[2];
    UCHAR MaximumPrefetchCeiling[2];
    UCHAR temp[7];
}MODE_PAGE8;

/*
  * Sense Data Format - Block Descriptor (DASD)
  */
typedef struct {
   ULONG NumberOfBlocks;
   UCHAR  DensityCode;
   USHORT BlockLengthHigh;
   UCHAR  BlockLengthLow;
}MODE_PAGE_BLKDESC;

/*
 * Sense Data Format - Mode Page Header
 */
typedef struct {
   UCHAR  DataLength;
   UCHAR  MediumType;
   UCHAR  Reserved;
   UCHAR  BlockDescLength;
}MODE_PAGE_HEADER;

typedef struct {
   MODE_PAGE_HEADER  hdr;
   MODE_PAGE_BLKDESC blkdesc;
   union {
	MODE_PAGE8 pg8;
	MODE_PAGE3 pg3;
	MODE_PAGE4 pg4;
	} pdata;
}MODE_PAGE_DATA;

typedef struct {
    ULONG LogicalBlockAddress;
    ULONG BytesPerBlock;
} READ_CAPACITY_DATA, *PREAD_CAPACITY_DATA;

// Used to store LUn specific information
typedef struct _LUN_EXTENSION
{
    //PVOID Address;			=> crlo
	PUCHAR	Address;
    ULONG Size;
//    SENSE_DATA SenseData;
} LUN_EXTENSION, *PLUN_EXTENSION;

typedef enum {
	MEDIA_IN_READER,
	NO_MEDIA_IN_READER,
	WRITE_PROTECT
} SENSETYPE;


void setscsiresult(struct scsi_cmnd *SCpnt,int ret);
#endif
